from collections.abc import Callable, Generator
from inspect import Signature

from .common import FOREVER, NOT_SET, Args
from .errors import CalledWithWrongValueError, InvalidArgumentsToCalledWithError


def _make_callable[**P, R](value: R) -> Callable[P, R]:
    return lambda *_, **__: value


def _make_callable_validate_args_partial[**P, R](
    args: tuple[object, ...], kwargs: dict[str, object], return_: R, sig: Signature
) -> Callable[P, R]:
    try:
        expected_arguments = sig.bind_partial(*args, **kwargs).arguments
    except TypeError as e:
        raise InvalidArgumentsToCalledWithError from e

    def validate_args(*args: P.args, **kwargs: P.kwargs) -> R:
        bound_arguments = sig.bind(*args, **kwargs).arguments
        for name, expected in expected_arguments.items():
            got = bound_arguments.get(name, NOT_SET)
            if got != expected:
                raise CalledWithWrongValueError(name=name, got=got, expected=expected)
        return return_

    return validate_args


def _make_callable_validate_args_full[**P, R](args: Args[P], return_: R, sig: Signature) -> Callable[P, R]:
    try:
        expected_arguments = sig.bind(*args.args, **args.kwargs).arguments
    except TypeError as e:
        raise InvalidArgumentsToCalledWithError from e

    def validate_args(*args: P.args, **kwargs: P.kwargs) -> R:
        bound_arguments = sig.bind(*args, **kwargs).arguments
        if bound_arguments != expected_arguments:
            for name in list(bound_arguments.keys()) + list(expected_arguments.keys()):
                expected = expected_arguments.get(name, NOT_SET)
                got = bound_arguments.get(name, NOT_SET)
                if expected != got:
                    raise CalledWithWrongValueError(name=name, got=got, expected=expected)
        return return_

    return validate_args


def _make_raise[**P, R, E: BaseException](error: E | type[E]) -> Callable[P, R]:
    def func(*_: P.args, **__: P.kwargs) -> R:
        raise error

    return func


def _make_generator[T](obj: T, times: int) -> Generator[T]:
    if times < -1 or times == 0:
        msg = f'Times should be either positive integer or FOREVER (-1), got: {times}'
        raise ValueError(msg)

    def generator() -> Generator[T]:
        if times == FOREVER:
            while True:
                yield obj
        else:
            for _ in range(times):
                yield obj

    return generator()


def return_impl[**P, R](value: R, *, times: int = 1) -> Generator[Callable[P, R]]:
    return _make_generator(_make_callable(value), times=times)


def raise_impl[**P, R, E: BaseException](error: E | type[E], *, times: int = 1) -> Generator[Callable[P, R]]:
    return _make_generator(_make_raise(error), times=times)


def called_with_partial_impl[**P, R](
    args: tuple[object, ...], kwargs: dict[str, object], return_: R, sig: Signature
) -> Generator[Callable[P, R]]:
    return _make_generator(_make_callable_validate_args_partial(args, kwargs, return_, sig), times=1)


def called_with_full_impl[**P, R](args: Args[P], return_: R, sig: Signature) -> Generator[Callable[P, R]]:
    return _make_generator(_make_callable_validate_args_full(args, return_, sig), times=1)



import functools
import inspect
from collections.abc import Awaitable, Callable, Generator
from inspect import Signature

from .common import ValidationConfig
from .errors import ValueIsNotSetError

type FakeMethod[**P, R] = Callable[P, R]
type Producer[**P, R] = Generator[FakeMethod[P, R]]


def _make_coro[**P, R](func: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> Awaitable[R]:
    async def inner() -> R:
        return func(*args, **kwargs)

    return inner()


class FakeMethodMember[**P, R]:
    def __init__(self, original_method: Callable[P, R], config: ValidationConfig) -> None:
        self.__producers: list[Producer[P, R]] = []
        self.__original_method = original_method
        self.__config = config
        functools.update_wrapper(self, original_method)

    def add_producer(self, producer: Producer[P, R]) -> None:
        self.__producers.append(producer)

    @property
    def signature(self) -> Signature:
        return inspect.signature(self.__original_method)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> object:
        sig = self.signature
        if self.__config.validate_call_arguments and sig:
            sig.bind(*args, **kwargs)

        producers = self.__producers
        while producers:
            producer = producers[0]
            try:
                func = next(producer)
            except StopIteration:
                producers.pop(0)
            else:
                if inspect.iscoroutinefunction(self.__original_method):
                    return _make_coro(func, *args, **kwargs)  # We do not want to have side effects before awaiting
                return func(*args, **kwargs)
        raise ValueIsNotSetError



from .common import FOREVER, NOT_SET, Args, ValidationConfig
from .errors import (
    CalledWithWrongValueError,
    FieldAccessedError,
    InvalidArgumentsToCalledWithError,
    MockingError,
    TestError,
    ValueIsNotSetError,
)
from .fake_method import Producer
from .fake_method_builder import FakeMethodBuilder
from .mocker import Mocker

__all__ = [
    'FOREVER',
    'NOT_SET',
    'Args',
    'CalledWithWrongValueError',
    'FakeMethodBuilder',
    'FieldAccessedError',
    'InvalidArgumentsToCalledWithError',
    'Mocker',
    'MockingError',
    'Producer',
    'TestError',
    'ValidationConfig',
    'ValueIsNotSetError',
]



from collections.abc import Awaitable
from typing import overload

from .common import Args
from .fake_method import FakeMethodMember
from .impl import called_with_full_impl, called_with_partial_impl, raise_impl, return_impl


class FakeMethodBuilder[**P, R]:
    def __init__(self, fake_method: FakeMethodMember[P, R | Awaitable[R]]) -> None:
        self.__fake_method = fake_method

    @overload
    def return_(self, /, *values: R) -> 'FakeMethodBuilder[P, R]': ...
    @overload
    def return_(self, value: R, /, *, times: int = 1) -> 'FakeMethodBuilder[P, R]': ...

    def return_(self, /, *values: R, times: int = 1) -> 'FakeMethodBuilder[P, R]':
        if not values:
            raise ValueError('No values specified')
        if len(values) > 1 and times != 1:
            raise ValueError('Cannot set more than 1 value and times at the same time')

        if len(values) > 1:
            for value in values:
                self.__fake_method.add_producer(return_impl(value))
        else:
            value = values[0]  # pyright: ignore[reportGeneralTypeIssues]
            self.__fake_method.add_producer(return_impl(value, times=times))
        return self

    def raise_[E: BaseException](self, error: E | type[E], *, times: int = 1) -> 'FakeMethodBuilder[P, R]':
        self.__fake_method.add_producer(raise_impl(error, times=times))
        return self

    def called_with(self, *args: object, return_: R, **kwargs: object) -> 'FakeMethodBuilder[P, R]':
        return self.called_with_partial(*args, return_=return_, **kwargs)

    def called_with_partial(self, *args: object, return_: R, **kwargs: object) -> 'FakeMethodBuilder[P, R]':
        self.__fake_method.add_producer(called_with_partial_impl(args, kwargs, return_, self.__fake_method.signature))
        return self

    def called_with_full(self, args: Args[P], return_: R) -> 'FakeMethodBuilder[P, R]':
        self.__fake_method.add_producer(called_with_full_impl(args, return_, self.__fake_method.signature))
        return self



from .common import NOT_SET


class MockingError(Exception):
    pass


class TestError(BaseException):
    pass


class InvalidArgumentsToCalledWithError(MockingError):
    pass


class ValueIsNotSetError(TestError):
    pass


class FieldAccessedError(TestError):
    pass


class CalledWithWrongValueError(TestError):
    name: str
    got: object | NOT_SET
    expected: object | NOT_SET

    def __init__(self, name: str, got: object | NOT_SET, expected: object, *args: object) -> None:
        self.name = name
        self.got = got
        self.expected = expected
        super().__init__(*args)



from typing import cast

from .common import ValidationConfig
from .errors import FieldAccessedError
from .fake_method import FakeMethodMember


class Mock[T]:
    def __init__(self, cls: type[T], config: ValidationConfig) -> None:
        self.__obj = object.__new__(cls)
        self.__config = config
        self.__fake_methods: dict[str, FakeMethodMember[..., object]] = {}

    def __getattribute__(self, name: str, /) -> object:
        if name == '__class__':
            return type(self.__obj)

        if name in ('_Mock__obj', '_Mock__config', '_Mock__fake_methods'):
            return super().__getattribute__(name)

        result = self.__obj.__getattribute__(name)
        config = self.__config

        if callable(result):
            return self.__fake_methods.setdefault(name, FakeMethodMember(result, config))

        if config.raise_on_field_access:
            raise FieldAccessedError
        return result


def create_mock[T](cls: type[T], config: ValidationConfig) -> T:
    mock = Mock(cls, config)
    return cast('T', mock)



from pydantic import BaseModel

FOREVER = -1


class NOT_SET:  # noqa: N801
    pass


class ValidationConfig(BaseModel):
    raise_on_field_access: bool = True  # Any callable is considered to be a non attribute
    validate_call_arguments: bool = True


class Args[**P]:
    def __init__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        self.args = args
        self.kwargs = kwargs



from collections.abc import Awaitable, Callable
from typing import overload

from .common import ValidationConfig
from .fake_method import FakeMethodMember
from .fake_method_builder import FakeMethodBuilder
from .mock import create_mock


class Mocker:
    def __init__(self, config: ValidationConfig | None = None) -> None:
        if config is None:
            config = ValidationConfig()

        self.config = config

    def mock[T](self, cls: type[T]) -> T:
        return create_mock(cls, self.config)

    @overload
    def when[**P, R](self, func: Callable[P, Awaitable[R]]) -> FakeMethodBuilder[P, R]: ...
    @overload
    def when[**P, R](self, func: Callable[P, R]) -> FakeMethodBuilder[P, R]: ...

    def when[**P, R](self, func: Callable[P, R | Awaitable[R]]) -> FakeMethodBuilder[P, R]:
        if not isinstance(func, FakeMethodMember):
            msg = f'Invalid argument to function `when`, expected class method, got: {type(func)}'
            raise TypeError(msg)

        return FakeMethodBuilder(func)
